/*
   The st.c file is a sub-stub file.  I just wanted to have all the detect code, etc in the
   mid level driver present and working.  If no one else volunteers for this, I'll
   do it - but it's low on my list of priorities.
*/
#include <linux/config.h>

#ifdef CONFIG_BLK_DEV_ST
#include "scsi.h"
#include "scsi_tape.h"
#include "st.h"

#define MAJOR_NR 9
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include "../blk.h" 
#include <asm/segment.h>

#define ST_TIMEOUT 1000
#define MAX_RETRIES 5
#define MIN(a, b)  ((a) < (b) ? (a) : (b))
/*
static const char RCSid[] = "$Header: /usr/src/linux/kernel/blk_drv/scsi/RCS/st.c,v 1.1 1992/04/24 18:01:50 root Exp root $";
*/

Scsi_Tape scsi_tapes[MAX_ST];
static int st_sizes[MAX_ST];
static int the_result;
static unsigned char sense_buffer[255];
static unsigned char buffer[128*512];
static unsigned char cmd[10];
int NR_ST=0;

static void st_sleep_done (int host, int result)
{

   the_result = result;
   wake_up( &scsi_tapes[0].waiting ); 
}

void decode_sns()
{
   static char *snstext[] = {
   "None","Recovered Error","Not Ready","Medium Error","Hardware Error",
   "Illegal Request","Unit Attention","Data Protect","Blank Check",
   "Key=E","Key=F","Filemark","End-Of-Medium","Incorrect Block Length",
   "14","15"};

   if (sense_buffer[0]!=0) {
      if ((sense_buffer[0] & 0x70) == 0x70) {
         if (sense_buffer[2] & 0x80) printk( "FMK ");
         if (sense_buffer[2] & 0x40) printk( "EOM ");
         if (sense_buffer[2] & 0x20) printk( "ILI ");
        printk( " %s\n", snstext[sense_buffer[2] & 0x0f]);
      } else {
      if (sense_buffer[0] < 15) printk( " %s\n", snstext[sense_buffer[0] & 0x0f]);
      else printk( "sns = %2x\n", sense_buffer[0]);
      }
   }
}

static void st_done (int host, int result)

{
   the_result = result;
}
static int scsi_tape_close(struct inode * inode, struct file * filp)
{
   int dev,major;
   int i,offset;
   static unsigned char cmd[10];

   
   dev = inode->i_rdev;
   major = dev >>8;
   dev = dev & 255; 

   if (major!=9)
      panic ("st.c: scsi_tape_open in scsi tape, but not major dev 9\n");

   if ( scsi_tapes[dev].rw == 2) {
      if (scsi_tapes[dev].dirty==1) {
         offset = filp->f_pos % 512;
         for (i=offset;i<512;i++) buffer[i] = 00;
         cmd[0] = WRITE_6;
         cmd[1] = 1;
         cmd[2] = 0;
         cmd[3] = 0;
         cmd[4] = 1;
         cmd[5] = 0;
         the_result = -1;
         scsi_do_cmd (scsi_tapes[dev].device->host_no,
                   scsi_tapes[dev].device->id,
                   (void *) cmd, buffer, 512, 
                   st_sleep_done, ST_TIMEOUT, sense_buffer, MAX_RETRIES);
         if (the_result <0) sleep_on( &scsi_tapes[0].waiting );
         filp->f_pos += offset;
      }
      for (i=0;i<10;i++) cmd[i]=0;
      cmd[0] = FILEMARK;
      cmd[4] = 1;
      the_result = -1;
      scsi_do_cmd( scsi_tapes[dev].device->host_no, 
                   scsi_tapes[dev].device->id,
                   (void *) cmd, (void *) buffer, 512,
                   st_sleep_done, ST_TIMEOUT, sense_buffer,
                   MAX_RETRIES);
      if (the_result <0) sleep_on( &scsi_tapes[0].waiting );
   printk("SCSI Tape buffer flushed, EOF marker wrote\n");
   }
   return 0;
}

static int scsi_tape_open(struct inode * inode, struct file * filp)
{
   int dev,major;
   int i;
   unsigned char buffer[513];
   unsigned char cmd[10];

   
   dev = inode->i_rdev;
   major = dev >>8;
   dev = dev & 255; 


   if (major!=9)
      panic ("st.c: scsi_tape_open in scsi tape, but not major dev 9\n");



   scsi_tapes[dev].dirty = 0;
   scsi_tapes[dev].rw = 0;


   for (i=0;i<10;i++) cmd[i]=0;
   for (i=0;i<12;i++) sense_buffer[i]=0;
   cmd[0] = TEST_UNIT_READY;
   the_result = -1;
   scsi_do_cmd( scsi_tapes[dev].device->host_no, 
                scsi_tapes[dev].device->id,
                (void *) cmd, (void *) buffer,
                512, st_sleep_done, ST_TIMEOUT, sense_buffer,
                MAX_RETRIES);

   if (the_result <0) sleep_on( &scsi_tapes[0].waiting );


   if (the_result != 0 ) {
      for (i=0;i<10;i++) cmd[i]=0;
      for (i=0;i<12;i++) sense_buffer[i]=0;
      cmd[0] = TEST_UNIT_READY;
      the_result = -1;
      scsi_do_cmd( scsi_tapes[dev].device->host_no, 
                   scsi_tapes[dev].device->id,
                   (void *) cmd, (void *) buffer,
                   512, st_sleep_done, ST_TIMEOUT, sense_buffer,
                   MAX_RETRIES);

      if (the_result <0) sleep_on( &scsi_tapes[0].waiting );
   }

   if (the_result == NO_TAPE ) {
      printk( " Tape %d says \"Feed Me Dummy!\"\n",dev);
      return -1;
   } else if (the_result != 0) {
      printk( " Tape %d dont want to talk reason = %d\n",dev,the_result);
      return -1;
   }

   for (i=0;i<10;i++) cmd[i]=0;
   for (i=0;i<12;i++) sense_buffer[i]=0;
   cmd[0] = MODE_SENSE;
   cmd[4] = 12;
   the_result = -1;
   scsi_do_cmd( scsi_tapes[dev].device->host_no, 
                scsi_tapes[dev].device->id,
                (void *) cmd, (void *) buffer,
                512, st_sleep_done, ST_TIMEOUT, sense_buffer,
                MAX_RETRIES);

   if (the_result <0) sleep_on( &scsi_tapes[0].waiting );

   if (buffer[2] & 0x80) printk( "tape is Write Protected\n");

   return 0;
}

int st_write(struct inode * inode, struct file * filp, char * buf, int count)
{

   int dev,major;
   int offset, block, total, do_count;
   static unsigned char cmd[10];
   char *b_point;

   dev = inode->i_rdev;
   major = dev >>8;
   dev = dev & 255;
 
   offset = filp->f_pos % 512;  
   block = (filp->f_pos - offset) / 512;
   total = count;

   cmd[0] = WRITE_6;
   cmd[1] = 1;
   cmd[2] = 0;
   cmd[3] = 0;
   cmd[4] = 1;
   cmd[5] = 0;

   scsi_tapes[dev].rw = 2;

   offset = filp->f_pos % 512;  
   block = (filp->f_pos - offset) / 512;
   b_point = buf;
   while((offset + count) >= 512) {
      do_count = 512 - offset;
      memcpy_fromfs(buffer+offset,b_point,do_count);


      the_result = -1;
      scsi_do_cmd (scsi_tapes[dev].device->host_no,
                scsi_tapes[dev].device->id,
                (void *) cmd, buffer, 512, 
                st_sleep_done, ST_TIMEOUT, sense_buffer, MAX_RETRIES);
      if (the_result <0) sleep_on( &scsi_tapes[0].waiting );
      filp->f_pos += do_count;
      b_point += do_count;
      count -= do_count;
      offset = 0;  
      scsi_tapes[dev].dirty = 0;
      block++;
   }
   if (count != 0) {
      scsi_tapes[dev].dirty = 1;
      memcpy_fromfs(buffer+offset,b_point,count);
      filp->f_pos += count;
      count = 0;
   }
   return( total);
}   

      
      
int st_read(struct inode * inode, struct file * filp, char * buf, int count)
{
   int dev,major;
   int do_count = 512;
   static unsigned char cmd[10];

   cmd[0] = READ_6;
   cmd[1] = 1;
   cmd[2] = 0;
   cmd[3] = 0;
   cmd[4] = 1;
   cmd[5] = 0;

   dev = inode->i_rdev;
   major = dev >>8;
   dev = dev & 255;

   scsi_tapes[dev].rw = 1;

   the_result = -1;
   scsi_do_cmd (scsi_tapes[dev].device->host_no,
             scsi_tapes[dev].device->id,
             (void *) cmd, buffer, do_count, 
             st_sleep_done, ST_TIMEOUT, sense_buffer, MAX_RETRIES);

   if (the_result <0) sleep_on( &scsi_tapes[0].waiting );

   if (! the_result) {
      memcpy_tofs(buf,buffer,do_count);
      filp->f_pos += do_count;
      return( do_count );
   } else if (((sense_buffer[0] & 0x70) == 0x70) && (sense_buffer[2] & 0x80)) { 
      printk( "EOF\n");
      return( -1 );
   } else {
      printk( "Got an Error\n");
      return(0);
   }
}

/*
	The st_init() function looks at all SCSI drives present, determines 
	their size, and reads partition	table entries for them.
*/




int st_ioctl(struct inode * inode,struct file * file, unsigned long cmd_in, unsigned long arg)
{
   int dev = inode->i_rdev;
   int i;


   dev = dev & 255; 
   switch (cmd_in) {
      case 1: 
         for (i=0;i<10;i++) cmd[i]=0;
         cmd[0] = ERASE;
         cmd[1] = 0x01;
         the_result = -1;
         printk( " Erasing Tape   : ");
         scsi_do_cmd( scsi_tapes[dev].device->host_no, 
                      scsi_tapes[dev].device->id,
                      (void *) cmd, (void *) buffer, 512,
                      st_sleep_done, ST_TIMEOUT, sense_buffer,
                      MAX_RETRIES);
         if (the_result <0) sleep_on( &scsi_tapes[0].waiting );
         break; 
      case 2: 
         for (i=0;i<10;i++) cmd[i]=0;
         cmd[0] = REZERO_UNIT;
         the_result = -1;
         printk( " Rezero Unit    : ");
         scsi_do_cmd( scsi_tapes[dev].device->host_no, 
                      scsi_tapes[dev].device->id,
                      (void *) cmd, (void *) buffer, 512,
                      st_sleep_done, ST_TIMEOUT, sense_buffer,
                      MAX_RETRIES);
         if (the_result <0) sleep_on( &scsi_tapes[0].waiting );
         break; 
      case 3: 
         for (i=0;i<10;i++) cmd[i]=0;
         cmd[0] = FILEMARK;
         cmd[4] = 1;
         the_result = -1;
         printk( " Write file mark: ");
         scsi_do_cmd( scsi_tapes[dev].device->host_no, 
                      scsi_tapes[dev].device->id,
                      (void *) cmd, (void *) buffer, 512,
                      st_sleep_done, ST_TIMEOUT, sense_buffer,
                      MAX_RETRIES);
         if (the_result <0) sleep_on( &scsi_tapes[0].waiting );
         break; 
      case 4: 
         for (i=0;i<10;i++) cmd[i]=0;
         cmd[0] = SPACE;
         cmd[1] = 1;
         cmd[4] = 1;
         the_result = -1;
         printk( " Read file mark : ");
         scsi_do_cmd( scsi_tapes[dev].device->host_no, 
                      scsi_tapes[dev].device->id,
                      (void *) cmd, (void *) buffer, 512,
                      st_sleep_done, ST_TIMEOUT, sense_buffer,
                      MAX_RETRIES);
         if (the_result <0) sleep_on( &scsi_tapes[0].waiting );
         break; 
      default:
         printk( "Unknow st_ioctl command \n");
         return -1;
   }
   return 0 ;
}
static struct file_operations st_fops = {
   NULL,            /* lseek - default */
   st_read,  /* read - general block-dev read */
   st_write, /* write - general block-dev write */
   NULL,            /* readdir - bad */
   NULL,            /* select */
   st_ioctl,        /* ioctl */
   scsi_tape_open,  /* open */
   scsi_tape_close  /* release */
};

void st_init(void)
{

   printk("Init tape\n");
   blkdev_fops[MAJOR_NR] = &st_fops;
}
#endif

   
